/* Sphere.d documents a Sphere class.
 * Joel C. Adams, Spring 2001
 * for C++ An Introduction to Computing 3/e
 */

#ifndef SPHERE
#define SPHERE

#include <cmath>      // pow()
#include <iostream>   // cerr, <<
using namespace std;

class Sphere
{
 public:
  /******** Instance Methods *********************************/
  /* Sphere constructor
   * POST: myRadius == 0 && myDensity == 0 && myWeight == 0.
   */
  Sphere();

  /* radius accessor
   * return: myRadius.
   */
  double getRadius();

  /* density accessor
   * return: myDensity.
   */
  double getDensity();

  /* weight accessor
   * return: myWeight.
   */
  double getWeight();

  /* radius-density mutator
   * params: newRadius and newDensity, both doubles.
   * PRE: newRadius > 0 && newDensity > 0.0.
   * POST: myRadius == newRadius && myDensity == newDensity &&
   *       myWeight == the weight for a sphere with myRadius, myDensity.
   */
 void setRadiusAndDensity(double newRadius, double newDensity);

  /* radius-weight mutator
   * params: newRadius and newWeight, both doubles.
   * PRE: newRadius > 0 && newWeight > 0.
   * POST: myRadius == newRadius && myWeight == newWeight &&
   *        myDensity == the density for a sphere with myRadius, myWeight.
   */
 void setRadiusAndWeight(double newRadius, double newWeight);


  /* density-weight mutator
   * params: newDensity and newWeight, both doubles.
   * PRE: newDensity > 0 && newWeight > 0.
   * POST: myDensity == newDensity && myWeight == newWeight &&
   *        myRadius == the radius of a sphere with myDensity, myWeight.
   */

  void setDensityAndWeight(double newDensity, double newWeight);


  /* output method
   * params: out, an ostream.
   * POST: a description of myself has been written to out.
   */
  void print(ostream& out);

  /* output operator function
   * params: out, an ostream; and aSphere, a Sphere.
   * return: out.
   * POST: aSphere has been printed to out.
   */
  friend ostream& operator<<(ostream& out, Sphere& aSphere);

  /* radius-and-density input
   * params: in, an istream.
   * PRE: in contains two doubles, r and d.
   * POST: r and d have been read from in &&
   *       myRadius == r && myDensity == d.
   */
  void   readRadiusAndDensity(istream& in);

  /* radius-and-weight input
   * params: in, an istream.
   * PRE: in contains two doubles, r and w.
   * POST: r and w have been read from in &&
   *       myRadius == r && myWeight == w.
   */
  void   readRadiusAndWeight(istream& in);

  /* density-and-weight input
   * params: in, an istream.
   * PRE: in contains two doubles, d and w.
   * POST: d and w have been read from in &&
   *       myDensity == d && myWeight == w.
   */
  void   readDensityAndWeight(istream& in);


  /******** Class Methods *********************************/
  /* sphere-radius class method
   * params: density, weight, both doubles.
   * PRE: density > 0 && weight > 0.
   * return: radius, for a given density and weight.
   */
  static double getRadius(double density, double weight);

  /* sphere-density class method
   * params: radius, weight, both doubles.
   * PRE: radius > 0 && weight > 0.
   * return: density, for a given radius and weight.
   */
  static double getDensity(double radius, double weight);

  /* sphere-weight class method
   * params: radius, density, both doubles.
   * PRE: radius > 0 && density > 0.
   * return: weight for a given radius and density.
   */
  static double getWeight(double radius, double density);

 private:
  double myRadius, myDensity, myWeight;
};



inline Sphere::Sphere()
{
  myRadius = myDensity = myWeight = 0.0;
}

inline double Sphere::getRadius() const
{
  return myRadius;
}

inline double Sphere::getDensity() const
{
  return myDensity;
}

inline double Sphere::getWeight() const
{
  return myWeight;
}

// Some useful constants ...
const double PI = 3.14159,
             ONE_THIRD = 1.0 / 3.0;

inline void Sphere::setRadiusAndDensity(double newRadius, double newDensity)
{
  if (newRadius > 0.0 && newDensity > 0.0) {
    myRadius = newRadius;
    myDensity = newDensity;
    myWeight = newDensity * 4.0 * PI * pow(newRadius, 3.0) / 3.0;
  }
  else {
    cerr << "setRadiusAndDensity(r,d): r,d must be non-negative" << endl;
  }
}

inline void Sphere::setRadiusAndWeight(double newRadius, double newWeight)
{
  if (newRadius > 0.0 && newWeight > 0.0) {
    myRadius = newRadius;
    myWeight = newWeight;
    myDensity = 3.0 * newWeight / (4.0 * PI * pow(newRadius, 3.0));
  }
  else {
    cerr << "setRadiusAndWeight(r,w): r,w must be non-negative" << endl;
  }
}

inline void Sphere::setDensityAndWeight(double newDensity, double newWeight)
{
  if (newDensity > 0.0 && newWeight > 0.0) {
    myDensity = newDensity;
    myWeight = newWeight;
    // pow(x, 1.0/3.0) == cube root of x
    myRadius = pow(3.0 * newWeight / (4.0 * PI * newDensity), ONE_THIRD);
  }
  else {
    cerr << "setDensityAndWeight(d,w): d,w must be non-negative" << endl;
  }
}

inline void Sphere::print(ostream & out) const
{
  out << "A sphere of radius " << getRadius()
      << ", density " << getDensity()
      << ", and weight " << getWeight();
}

inline ostream& operator<<(ostream& out, const Sphere& aSphere)
{
  aSphere.print(out);
  return out;
}

inline void Sphere::readRadiusAndDensity(istream& in)
{
  double radius, density;
  in >> radius >> density;
  setRadiusAndDensity(radius, density);
}

inline void Sphere::readRadiusAndWeight(istream& in)
{
  double radius, weight;
  in >> radius >> weight;
  setRadiusAndWeight(radius, weight);
}

inline void Sphere::readDensityAndWeight(istream& in)
{
  double density, weight;
  in >> density >> weight;
  setDensityAndWeight(density, weight);
}

inline double Sphere::getRadius(double density, double weight)
{
  if (density > 0.0 && weight > 0.0)
    // pow(x, 1.0/3.0) == cube root of x
    return pow(3.0 * weight / (4.0 * PI * density), ONE_THIRD);
  else
    return 0.0;
}

inline double Sphere::getDensity(double radius, double weight)
{
  if (radius > 0.0 && weight > 0.0)
    return 3.0 * weight / (4.0 * PI * pow(radius, 3.0));
  else
    return 0.0;
}

inline double Sphere::getWeight(double radius, double density)
{
  if (radius > 0.0 && density > 0.0)
    return density * 4.0 * PI * pow(radius, 3.0) / 3.0;
  else
    return 0.0;
}

#endif
